from asdl import pybase
from typing import Optional, List, Tuple, Dict, Any, cast, TYPE_CHECKING

if TYPE_CHECKING:
  from _devbuild.gen.syntax_asdl import loc_t, Token, expr_t, command_t, DoubleQuoted, re_t, proc_sig_t, LiteralBlock, Func
  from _devbuild.gen.runtime_asdl import Cell


from asdl import runtime  # For runtime.NO_SPID
from asdl.runtime import NewRecord, NewLeaf
from _devbuild.gen.hnode_asdl import color_e, hnode, hnode_e, hnode_t, Field

class y_lvalue_e(object):
  Local = 66
  Container = 2

_y_lvalue_str = {
  2: 'Container',
  66: 'Local',
}

def y_lvalue_str(tag, dot=True):
  # type: (int, bool) -> str
  v = _y_lvalue_str[tag]
  if dot:
    return "y_lvalue.%s" % v
  else:
    return v

class y_lvalue_t(pybase.CompoundObj):
  def tag(self):
    # type: () -> int
    return self._type_tag
  pass

class y_lvalue(object):
  class Container(y_lvalue_t):
    _type_tag = 2
    __slots__ = ('obj', 'index')
  
    def __init__(self, obj, index):
      # type: (value_t, value_t) -> None
      self.obj = obj
      self.index = index
  
    @staticmethod
    def CreateNull(alloc_lists=False):
      # type: () -> y_lvalue.Container
      return y_lvalue.Container(cast(value_t, None), cast(value_t, None))
  
    def PrettyTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('y_lvalue.Container')
      L = out_node.fields
  
      assert self.obj is not None
      x0 = self.obj.PrettyTree()
      L.append(Field('obj', x0))
  
      assert self.index is not None
      x1 = self.index.PrettyTree()
      L.append(Field('index', x1))
  
      return out_node
  
    def _AbbreviatedTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('y_lvalue.Container')
      L = out_node.fields
      assert self.obj is not None
      x0 = self.obj.AbbreviatedTree()
      L.append(Field('obj', x0))
  
      assert self.index is not None
      x1 = self.index.AbbreviatedTree()
      L.append(Field('index', x1))
  
      return out_node
  
    def AbbreviatedTree(self):
      # type: () -> hnode_t
      return self._AbbreviatedTree()
  
  pass

class sh_lvalue_e(object):
  Var = 66
  Indexed = 2
  Keyed = 3

_sh_lvalue_str = {
  2: 'Indexed',
  3: 'Keyed',
  66: 'Var',
}

def sh_lvalue_str(tag, dot=True):
  # type: (int, bool) -> str
  v = _sh_lvalue_str[tag]
  if dot:
    return "sh_lvalue.%s" % v
  else:
    return v

class sh_lvalue_t(pybase.CompoundObj):
  def tag(self):
    # type: () -> int
    return self._type_tag
  pass

class sh_lvalue(object):
  class Indexed(sh_lvalue_t):
    _type_tag = 2
    __slots__ = ('name', 'index', 'blame_loc')
  
    def __init__(self, name, index, blame_loc):
      # type: (str, int, loc_t) -> None
      self.name = name
      self.index = index
      self.blame_loc = blame_loc
  
    @staticmethod
    def CreateNull(alloc_lists=False):
      # type: () -> sh_lvalue.Indexed
      return sh_lvalue.Indexed('', -1, cast(loc_t, None))
  
    def PrettyTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('sh_lvalue.Indexed')
      L = out_node.fields
  
      x0 = NewLeaf(self.name, color_e.StringConst)
      L.append(Field('name', x0))
  
      x1 = hnode.Leaf(str(self.index), color_e.OtherConst)
      L.append(Field('index', x1))
  
      assert self.blame_loc is not None
      x2 = self.blame_loc.PrettyTree()
      L.append(Field('blame_loc', x2))
  
      return out_node
  
    def _AbbreviatedTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('sh_lvalue.Indexed')
      L = out_node.fields
      x0 = NewLeaf(self.name, color_e.StringConst)
      L.append(Field('name', x0))
  
      x1 = hnode.Leaf(str(self.index), color_e.OtherConst)
      L.append(Field('index', x1))
  
      assert self.blame_loc is not None
      x2 = self.blame_loc.AbbreviatedTree()
      L.append(Field('blame_loc', x2))
  
      return out_node
  
    def AbbreviatedTree(self):
      # type: () -> hnode_t
      return self._AbbreviatedTree()
  
  class Keyed(sh_lvalue_t):
    _type_tag = 3
    __slots__ = ('name', 'key', 'blame_loc')
  
    def __init__(self, name, key, blame_loc):
      # type: (str, str, loc_t) -> None
      self.name = name
      self.key = key
      self.blame_loc = blame_loc
  
    @staticmethod
    def CreateNull(alloc_lists=False):
      # type: () -> sh_lvalue.Keyed
      return sh_lvalue.Keyed('', '', cast(loc_t, None))
  
    def PrettyTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('sh_lvalue.Keyed')
      L = out_node.fields
  
      x0 = NewLeaf(self.name, color_e.StringConst)
      L.append(Field('name', x0))
  
      x1 = NewLeaf(self.key, color_e.StringConst)
      L.append(Field('key', x1))
  
      assert self.blame_loc is not None
      x2 = self.blame_loc.PrettyTree()
      L.append(Field('blame_loc', x2))
  
      return out_node
  
    def _AbbreviatedTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('sh_lvalue.Keyed')
      L = out_node.fields
      x0 = NewLeaf(self.name, color_e.StringConst)
      L.append(Field('name', x0))
  
      x1 = NewLeaf(self.key, color_e.StringConst)
      L.append(Field('key', x1))
  
      assert self.blame_loc is not None
      x2 = self.blame_loc.AbbreviatedTree()
      L.append(Field('blame_loc', x2))
  
      return out_node
  
    def AbbreviatedTree(self):
      # type: () -> hnode_t
      return self._AbbreviatedTree()
  
  pass

class value_e(object):
  Undef = 1
  Str = 2
  BashArray = 3
  BashAssoc = 4
  Null = 5
  Bool = 6
  Int = 7
  Float = 8
  List = 9
  Dict = 10
  Eggex = 11
  Expr = 12
  Template = 13
  Command = 14
  Block = 15
  Place = 16
  Module = 17
  IO = 18
  BuiltinFunc = 19
  BoundFunc = 20
  Proc = 21
  Func = 22
  Slice = 23
  Range = 24

_value_str = {
  1: 'Undef',
  2: 'Str',
  3: 'BashArray',
  4: 'BashAssoc',
  5: 'Null',
  6: 'Bool',
  7: 'Int',
  8: 'Float',
  9: 'List',
  10: 'Dict',
  11: 'Eggex',
  12: 'Expr',
  13: 'Template',
  14: 'Command',
  15: 'Block',
  16: 'Place',
  17: 'Module',
  18: 'IO',
  19: 'BuiltinFunc',
  20: 'BoundFunc',
  21: 'Proc',
  22: 'Func',
  23: 'Slice',
  24: 'Range',
}

def value_str(tag, dot=True):
  # type: (int, bool) -> str
  v = _value_str[tag]
  if dot:
    return "value.%s" % v
  else:
    return v

class value_t(pybase.CompoundObj):
  def tag(self):
    # type: () -> int
    return self._type_tag
  pass

class value__Undef(value_t):
  _type_tag = 1
  __slots__ = ()

  def __init__(self, ):
    # type: () -> None
    pass

  def PrettyTree(self):
    # type: () -> hnode_t
    out_node = NewRecord('value__Undef')
    L = out_node.fields

    return out_node

  def _AbbreviatedTree(self):
    # type: () -> hnode_t
    out_node = NewRecord('value__Undef')
    L = out_node.fields
    return out_node

  def AbbreviatedTree(self):
    # type: () -> hnode_t
    return self._AbbreviatedTree()

class value__Null(value_t):
  _type_tag = 5
  __slots__ = ()

  def __init__(self, ):
    # type: () -> None
    pass

  def PrettyTree(self):
    # type: () -> hnode_t
    out_node = NewRecord('value__Null')
    L = out_node.fields

    return out_node

  def _AbbreviatedTree(self):
    # type: () -> hnode_t
    out_node = NewRecord('value__Null')
    L = out_node.fields
    return out_node

  def AbbreviatedTree(self):
    # type: () -> hnode_t
    return self._AbbreviatedTree()

class value(object):
  Undef = value__Undef()
  
  class Str(value_t):
    _type_tag = 2
    __slots__ = ('s',)
  
    def __init__(self, s):
      # type: (str) -> None
      self.s = s
  
    @staticmethod
    def CreateNull(alloc_lists=False):
      # type: () -> value.Str
      return value.Str('')
  
    def PrettyTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Str')
      L = out_node.fields
  
      x0 = NewLeaf(self.s, color_e.StringConst)
      L.append(Field('s', x0))
  
      return out_node
  
    def _AbbreviatedTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Str')
      L = out_node.fields
      x0 = NewLeaf(self.s, color_e.StringConst)
      L.append(Field('s', x0))
  
      return out_node
  
    def AbbreviatedTree(self):
      # type: () -> hnode_t
      return self._AbbreviatedTree()
  
  class BashArray(value_t):
    _type_tag = 3
    __slots__ = ('strs',)
  
    def __init__(self, strs):
      # type: (List[str]) -> None
      self.strs = strs
  
    @staticmethod
    def CreateNull(alloc_lists=False):
      # type: () -> value.BashArray
      return value.BashArray([] if alloc_lists else cast('List[str]', None))
  
    def PrettyTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.BashArray')
      L = out_node.fields
  
      if self.strs is not None:  # List
        x0 = hnode.Array([])
        for i0 in self.strs:
          x0.children.append(NewLeaf(i0, color_e.StringConst))
        L.append(Field('strs', x0))
  
      return out_node
  
    def _AbbreviatedTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.BashArray')
      L = out_node.fields
      if self.strs is not None:  # List
        x0 = hnode.Array([])
        for i0 in self.strs:
          x0.children.append(NewLeaf(i0, color_e.StringConst))
        L.append(Field('strs', x0))
  
      return out_node
  
    def AbbreviatedTree(self):
      # type: () -> hnode_t
      return self._AbbreviatedTree()
  
  class BashAssoc(value_t):
    _type_tag = 4
    __slots__ = ('d',)
  
    def __init__(self, d):
      # type: (Dict[str, str]) -> None
      self.d = d
  
    @staticmethod
    def CreateNull(alloc_lists=False):
      # type: () -> value.BashAssoc
      return value.BashAssoc(cast('Dict[str, str]', None))
  
    def PrettyTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.BashAssoc')
      L = out_node.fields
  
      if self.d is not None:  # Dict
        m = hnode.Leaf("Dict", color_e.OtherConst)
        x0 = hnode.Array([m])
        for k0, v0 in self.d.iteritems():
          x0.children.append(NewLeaf(k0, color_e.StringConst))
          x0.children.append(NewLeaf(v0, color_e.StringConst))
        L.append(Field('d', x0))
  
      return out_node
  
    def _AbbreviatedTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.BashAssoc')
      L = out_node.fields
      if self.d is not None:  # Dict
        m = hnode.Leaf("Dict", color_e.OtherConst)
        x0 = hnode.Array([m])
        for k0, v0 in self.d.iteritems():
          x0.children.append(NewLeaf(k0, color_e.StringConst))
          x0.children.append(NewLeaf(v0, color_e.StringConst))
        L.append(Field('d', x0))
  
      return out_node
  
    def AbbreviatedTree(self):
      # type: () -> hnode_t
      return self._AbbreviatedTree()
  
  Null = value__Null()
  
  class Bool(value_t):
    _type_tag = 6
    __slots__ = ('b',)
  
    def __init__(self, b):
      # type: (bool) -> None
      self.b = b
  
    @staticmethod
    def CreateNull(alloc_lists=False):
      # type: () -> value.Bool
      return value.Bool(False)
  
    def PrettyTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Bool')
      L = out_node.fields
  
      x0 = hnode.Leaf('T' if self.b else 'F', color_e.OtherConst)
      L.append(Field('b', x0))
  
      return out_node
  
    def _AbbreviatedTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Bool')
      L = out_node.fields
      x0 = hnode.Leaf('T' if self.b else 'F', color_e.OtherConst)
      L.append(Field('b', x0))
  
      return out_node
  
    def AbbreviatedTree(self):
      # type: () -> hnode_t
      return self._AbbreviatedTree()
  
  class Int(value_t):
    _type_tag = 7
    __slots__ = ('i',)
  
    def __init__(self, i):
      # type: (int) -> None
      self.i = i
  
    @staticmethod
    def CreateNull(alloc_lists=False):
      # type: () -> value.Int
      return value.Int(-1)
  
    def PrettyTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Int')
      L = out_node.fields
  
      x0 = hnode.Leaf(str(self.i), color_e.OtherConst)
      L.append(Field('i', x0))
  
      return out_node
  
    def _AbbreviatedTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Int')
      L = out_node.fields
      x0 = hnode.Leaf(str(self.i), color_e.OtherConst)
      L.append(Field('i', x0))
  
      return out_node
  
    def AbbreviatedTree(self):
      # type: () -> hnode_t
      return self._AbbreviatedTree()
  
  class Float(value_t):
    _type_tag = 8
    __slots__ = ('f',)
  
    def __init__(self, f):
      # type: (float) -> None
      self.f = f
  
    @staticmethod
    def CreateNull(alloc_lists=False):
      # type: () -> value.Float
      return value.Float(0.0)
  
    def PrettyTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Float')
      L = out_node.fields
  
      x0 = hnode.Leaf(str(self.f), color_e.OtherConst)
      L.append(Field('f', x0))
  
      return out_node
  
    def _AbbreviatedTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Float')
      L = out_node.fields
      x0 = hnode.Leaf(str(self.f), color_e.OtherConst)
      L.append(Field('f', x0))
  
      return out_node
  
    def AbbreviatedTree(self):
      # type: () -> hnode_t
      return self._AbbreviatedTree()
  
  class List(value_t):
    _type_tag = 9
    __slots__ = ('items',)
  
    def __init__(self, items):
      # type: (List[value_t]) -> None
      self.items = items
  
    @staticmethod
    def CreateNull(alloc_lists=False):
      # type: () -> value.List
      return value.List([] if alloc_lists else cast('List[value_t]', None))
  
    def PrettyTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.List')
      L = out_node.fields
  
      if self.items is not None:  # List
        x0 = hnode.Array([])
        for i0 in self.items:
          x0.children.append(i0.PrettyTree())
        L.append(Field('items', x0))
  
      return out_node
  
    def _AbbreviatedTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.List')
      L = out_node.fields
      if self.items is not None:  # List
        x0 = hnode.Array([])
        for i0 in self.items:
          x0.children.append(i0.AbbreviatedTree())
        L.append(Field('items', x0))
  
      return out_node
  
    def AbbreviatedTree(self):
      # type: () -> hnode_t
      return self._AbbreviatedTree()
  
  class Dict(value_t):
    _type_tag = 10
    __slots__ = ('d',)
  
    def __init__(self, d):
      # type: (Dict[str, value_t]) -> None
      self.d = d
  
    @staticmethod
    def CreateNull(alloc_lists=False):
      # type: () -> value.Dict
      return value.Dict(cast('Dict[str, value_t]', None))
  
    def PrettyTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Dict')
      L = out_node.fields
  
      if self.d is not None:  # Dict
        m = hnode.Leaf("Dict", color_e.OtherConst)
        x0 = hnode.Array([m])
        for k0, v0 in self.d.iteritems():
          x0.children.append(NewLeaf(k0, color_e.StringConst))
          x0.children.append(v0.PrettyTree())
        L.append(Field('d', x0))
  
      return out_node
  
    def _AbbreviatedTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Dict')
      L = out_node.fields
      if self.d is not None:  # Dict
        m = hnode.Leaf("Dict", color_e.OtherConst)
        x0 = hnode.Array([m])
        for k0, v0 in self.d.iteritems():
          x0.children.append(NewLeaf(k0, color_e.StringConst))
          x0.children.append(v0.AbbreviatedTree())
        L.append(Field('d', x0))
  
      return out_node
  
    def AbbreviatedTree(self):
      # type: () -> hnode_t
      return self._AbbreviatedTree()
  
  class Eggex(value_t):
    _type_tag = 11
    __slots__ = ('expr', 'as_ere')
  
    def __init__(self, expr, as_ere):
      # type: (re_t, Optional[str]) -> None
      self.expr = expr
      self.as_ere = as_ere
  
    @staticmethod
    def CreateNull(alloc_lists=False):
      # type: () -> value.Eggex
      return value.Eggex(cast(re_t, None), cast('Optional[str]', None))
  
    def PrettyTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Eggex')
      L = out_node.fields
  
      assert self.expr is not None
      x0 = self.expr.PrettyTree()
      L.append(Field('expr', x0))
  
      if self.as_ere is not None:  # Optional
        x1 = NewLeaf(self.as_ere, color_e.StringConst)
        L.append(Field('as_ere', x1))
  
      return out_node
  
    def _AbbreviatedTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Eggex')
      L = out_node.fields
      assert self.expr is not None
      x0 = self.expr.AbbreviatedTree()
      L.append(Field('expr', x0))
  
      if self.as_ere is not None:  # Optional
        x1 = NewLeaf(self.as_ere, color_e.StringConst)
        L.append(Field('as_ere', x1))
  
      return out_node
  
    def AbbreviatedTree(self):
      # type: () -> hnode_t
      return self._AbbreviatedTree()
  
  class Expr(value_t):
    _type_tag = 12
    __slots__ = ('e',)
  
    def __init__(self, e):
      # type: (expr_t) -> None
      self.e = e
  
    @staticmethod
    def CreateNull(alloc_lists=False):
      # type: () -> value.Expr
      return value.Expr(cast(expr_t, None))
  
    def PrettyTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Expr')
      L = out_node.fields
  
      assert self.e is not None
      x0 = self.e.PrettyTree()
      L.append(Field('e', x0))
  
      return out_node
  
    def _AbbreviatedTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Expr')
      L = out_node.fields
      assert self.e is not None
      x0 = self.e.AbbreviatedTree()
      L.append(Field('e', x0))
  
      return out_node
  
    def AbbreviatedTree(self):
      # type: () -> hnode_t
      return self._AbbreviatedTree()
  
  class Template(value_t):
    _type_tag = 13
    __slots__ = ('dq',)
  
    def __init__(self, dq):
      # type: (DoubleQuoted) -> None
      self.dq = dq
  
    @staticmethod
    def CreateNull(alloc_lists=False):
      # type: () -> value.Template
      return value.Template(cast(DoubleQuoted, None))
  
    def PrettyTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Template')
      L = out_node.fields
  
      assert self.dq is not None
      x0 = self.dq.PrettyTree()
      L.append(Field('dq', x0))
  
      return out_node
  
    def _AbbreviatedTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Template')
      L = out_node.fields
      assert self.dq is not None
      x0 = self.dq.AbbreviatedTree()
      L.append(Field('dq', x0))
  
      return out_node
  
    def AbbreviatedTree(self):
      # type: () -> hnode_t
      return self._AbbreviatedTree()
  
  class Command(value_t):
    _type_tag = 14
    __slots__ = ('c',)
  
    def __init__(self, c):
      # type: (command_t) -> None
      self.c = c
  
    @staticmethod
    def CreateNull(alloc_lists=False):
      # type: () -> value.Command
      return value.Command(cast(command_t, None))
  
    def PrettyTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Command')
      L = out_node.fields
  
      assert self.c is not None
      x0 = self.c.PrettyTree()
      L.append(Field('c', x0))
  
      return out_node
  
    def _AbbreviatedTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Command')
      L = out_node.fields
      assert self.c is not None
      x0 = self.c.AbbreviatedTree()
      L.append(Field('c', x0))
  
      return out_node
  
    def AbbreviatedTree(self):
      # type: () -> hnode_t
      return self._AbbreviatedTree()
  
  class Block(value_t):
    _type_tag = 15
    __slots__ = ('block',)
  
    def __init__(self, block):
      # type: (LiteralBlock) -> None
      self.block = block
  
    @staticmethod
    def CreateNull(alloc_lists=False):
      # type: () -> value.Block
      return value.Block(cast(LiteralBlock, None))
  
    def PrettyTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Block')
      L = out_node.fields
  
      assert self.block is not None
      x0 = self.block.PrettyTree()
      L.append(Field('block', x0))
  
      return out_node
  
    def _AbbreviatedTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Block')
      L = out_node.fields
      assert self.block is not None
      x0 = self.block.AbbreviatedTree()
      L.append(Field('block', x0))
  
      return out_node
  
    def AbbreviatedTree(self):
      # type: () -> hnode_t
      return self._AbbreviatedTree()
  
  class Place(value_t):
    _type_tag = 16
    __slots__ = ('lval', 'frame')
  
    def __init__(self, lval, frame):
      # type: (y_lvalue_t, Dict[str, Cell]) -> None
      self.lval = lval
      self.frame = frame
  
    @staticmethod
    def CreateNull(alloc_lists=False):
      # type: () -> value.Place
      return value.Place(cast(y_lvalue_t, None), cast('Dict[str, Cell]', None))
  
    def PrettyTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Place')
      L = out_node.fields
  
      assert self.lval is not None
      x0 = self.lval.PrettyTree()
      L.append(Field('lval', x0))
  
      if self.frame is not None:  # Dict
        m = hnode.Leaf("Dict", color_e.OtherConst)
        x1 = hnode.Array([m])
        for k1, v1 in self.frame.iteritems():
          x1.children.append(NewLeaf(k1, color_e.StringConst))
          x1.children.append(v1.PrettyTree())
        L.append(Field('frame', x1))
  
      return out_node
  
    def _AbbreviatedTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Place')
      L = out_node.fields
      assert self.lval is not None
      x0 = self.lval.AbbreviatedTree()
      L.append(Field('lval', x0))
  
      if self.frame is not None:  # Dict
        m = hnode.Leaf("Dict", color_e.OtherConst)
        x1 = hnode.Array([m])
        for k1, v1 in self.frame.iteritems():
          x1.children.append(NewLeaf(k1, color_e.StringConst))
          x1.children.append(v1.AbbreviatedTree())
        L.append(Field('frame', x1))
  
      return out_node
  
    def AbbreviatedTree(self):
      # type: () -> hnode_t
      return self._AbbreviatedTree()
  
  class Module(value_t):
    _type_tag = 17
    __slots__ = ('defs',)
  
    def __init__(self, defs):
      # type: (Dict[str, value_t]) -> None
      self.defs = defs
  
    @staticmethod
    def CreateNull(alloc_lists=False):
      # type: () -> value.Module
      return value.Module(cast('Dict[str, value_t]', None))
  
    def PrettyTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Module')
      L = out_node.fields
  
      if self.defs is not None:  # Dict
        m = hnode.Leaf("Dict", color_e.OtherConst)
        x0 = hnode.Array([m])
        for k0, v0 in self.defs.iteritems():
          x0.children.append(NewLeaf(k0, color_e.StringConst))
          x0.children.append(v0.PrettyTree())
        L.append(Field('defs', x0))
  
      return out_node
  
    def _AbbreviatedTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Module')
      L = out_node.fields
      if self.defs is not None:  # Dict
        m = hnode.Leaf("Dict", color_e.OtherConst)
        x0 = hnode.Array([m])
        for k0, v0 in self.defs.iteritems():
          x0.children.append(NewLeaf(k0, color_e.StringConst))
          x0.children.append(v0.AbbreviatedTree())
        L.append(Field('defs', x0))
  
      return out_node
  
    def AbbreviatedTree(self):
      # type: () -> hnode_t
      return self._AbbreviatedTree()
  
  class IO(value_t):
    _type_tag = 18
    __slots__ = ('cmd_ev', 'prompt_ev')
  
    def __init__(self, cmd_ev, prompt_ev):
      # type: (Any, Any) -> None
      self.cmd_ev = cmd_ev
      self.prompt_ev = prompt_ev
  
    @staticmethod
    def CreateNull(alloc_lists=False):
      # type: () -> value.IO
      return value.IO(cast(Any, None), cast(Any, None))
  
    def PrettyTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.IO')
      L = out_node.fields
  
      x0 = hnode.External(self.cmd_ev)
      L.append(Field('cmd_ev', x0))
  
      x1 = hnode.External(self.prompt_ev)
      L.append(Field('prompt_ev', x1))
  
      return out_node
  
    def _AbbreviatedTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.IO')
      L = out_node.fields
      x0 = hnode.External(self.cmd_ev)
      L.append(Field('cmd_ev', x0))
  
      x1 = hnode.External(self.prompt_ev)
      L.append(Field('prompt_ev', x1))
  
      return out_node
  
    def AbbreviatedTree(self):
      # type: () -> hnode_t
      return self._AbbreviatedTree()
  
  class BuiltinFunc(value_t):
    _type_tag = 19
    __slots__ = ('callable',)
  
    def __init__(self, callable):
      # type: (Any) -> None
      self.callable = callable
  
    @staticmethod
    def CreateNull(alloc_lists=False):
      # type: () -> value.BuiltinFunc
      return value.BuiltinFunc(cast(Any, None))
  
    def PrettyTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.BuiltinFunc')
      L = out_node.fields
  
      x0 = hnode.External(self.callable)
      L.append(Field('callable', x0))
  
      return out_node
  
    def _AbbreviatedTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.BuiltinFunc')
      L = out_node.fields
      x0 = hnode.External(self.callable)
      L.append(Field('callable', x0))
  
      return out_node
  
    def AbbreviatedTree(self):
      # type: () -> hnode_t
      return self._AbbreviatedTree()
  
  class BoundFunc(value_t):
    _type_tag = 20
    __slots__ = ('me', 'func')
  
    def __init__(self, me, func):
      # type: (value_t, value_t) -> None
      self.me = me
      self.func = func
  
    @staticmethod
    def CreateNull(alloc_lists=False):
      # type: () -> value.BoundFunc
      return value.BoundFunc(cast(value_t, None), cast(value_t, None))
  
    def PrettyTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.BoundFunc')
      L = out_node.fields
  
      assert self.me is not None
      x0 = self.me.PrettyTree()
      L.append(Field('me', x0))
  
      assert self.func is not None
      x1 = self.func.PrettyTree()
      L.append(Field('func', x1))
  
      return out_node
  
    def _AbbreviatedTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.BoundFunc')
      L = out_node.fields
      assert self.me is not None
      x0 = self.me.AbbreviatedTree()
      L.append(Field('me', x0))
  
      assert self.func is not None
      x1 = self.func.AbbreviatedTree()
      L.append(Field('func', x1))
  
      return out_node
  
    def AbbreviatedTree(self):
      # type: () -> hnode_t
      return self._AbbreviatedTree()
  
  class Proc(value_t):
    _type_tag = 21
    __slots__ = ('name', 'name_tok', 'sig', 'body', 'defaults', 'dynamic_scope')
  
    def __init__(self, name, name_tok, sig, body, defaults, dynamic_scope):
      # type: (str, Token, proc_sig_t, command_t, Optional[ProcDefaults], bool) -> None
      self.name = name
      self.name_tok = name_tok
      self.sig = sig
      self.body = body
      self.defaults = defaults
      self.dynamic_scope = dynamic_scope
  
    @staticmethod
    def CreateNull(alloc_lists=False):
      # type: () -> value.Proc
      return value.Proc('', cast(Token, None), cast(proc_sig_t, None), cast(command_t, None), cast('Optional[ProcDefaults]', None), False)
  
    def PrettyTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Proc')
      L = out_node.fields
  
      x0 = NewLeaf(self.name, color_e.StringConst)
      L.append(Field('name', x0))
  
      assert self.name_tok is not None
      x1 = self.name_tok.PrettyTree()
      L.append(Field('name_tok', x1))
  
      assert self.sig is not None
      x2 = self.sig.PrettyTree()
      L.append(Field('sig', x2))
  
      assert self.body is not None
      x3 = self.body.PrettyTree()
      L.append(Field('body', x3))
  
      if self.defaults is not None:  # Optional
        x4 = self.defaults.PrettyTree()
        L.append(Field('defaults', x4))
  
      x5 = hnode.Leaf('T' if self.dynamic_scope else 'F', color_e.OtherConst)
      L.append(Field('dynamic_scope', x5))
  
      return out_node
  
    def _AbbreviatedTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Proc')
      L = out_node.fields
      x0 = NewLeaf(self.name, color_e.StringConst)
      L.append(Field('name', x0))
  
      assert self.name_tok is not None
      x1 = self.name_tok.AbbreviatedTree()
      L.append(Field('name_tok', x1))
  
      assert self.sig is not None
      x2 = self.sig.AbbreviatedTree()
      L.append(Field('sig', x2))
  
      assert self.body is not None
      x3 = self.body.AbbreviatedTree()
      L.append(Field('body', x3))
  
      if self.defaults is not None:  # Optional
        x4 = self.defaults.AbbreviatedTree()
        L.append(Field('defaults', x4))
  
      x5 = hnode.Leaf('T' if self.dynamic_scope else 'F', color_e.OtherConst)
      L.append(Field('dynamic_scope', x5))
  
      return out_node
  
    def AbbreviatedTree(self):
      # type: () -> hnode_t
      return self._AbbreviatedTree()
  
  class Func(value_t):
    _type_tag = 22
    __slots__ = ('name', 'parsed', 'pos_defaults', 'named_defaults', 'module_')
  
    def __init__(self, name, parsed, pos_defaults, named_defaults, module_):
      # type: (str, Func, List[value_t], Dict[str, value_t], Optional[Dict[str, Cell]]) -> None
      self.name = name
      self.parsed = parsed
      self.pos_defaults = pos_defaults
      self.named_defaults = named_defaults
      self.module_ = module_
  
    @staticmethod
    def CreateNull(alloc_lists=False):
      # type: () -> value.Func
      return value.Func('', cast(Func, None), [] if alloc_lists else cast('List[value_t]', None), cast('Dict[str, value_t]', None), cast('Optional[Dict[str, Cell]]', None))
  
    def PrettyTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Func')
      L = out_node.fields
  
      x0 = NewLeaf(self.name, color_e.StringConst)
      L.append(Field('name', x0))
  
      assert self.parsed is not None
      x1 = self.parsed.PrettyTree()
      L.append(Field('parsed', x1))
  
      if self.pos_defaults is not None:  # List
        x2 = hnode.Array([])
        for i2 in self.pos_defaults:
          x2.children.append(i2.PrettyTree())
        L.append(Field('pos_defaults', x2))
  
      if self.named_defaults is not None:  # Dict
        m = hnode.Leaf("Dict", color_e.OtherConst)
        x3 = hnode.Array([m])
        for k3, v3 in self.named_defaults.iteritems():
          x3.children.append(NewLeaf(k3, color_e.StringConst))
          x3.children.append(v3.PrettyTree())
        L.append(Field('named_defaults', x3))
  
      if self.module_ is not None:  # Dict
        m = hnode.Leaf("Dict", color_e.OtherConst)
        x4 = hnode.Array([m])
        for k4, v4 in self.module_.iteritems():
          x4.children.append(NewLeaf(k4, color_e.StringConst))
          x4.children.append(v4.PrettyTree())
        L.append(Field('module_', x4))
  
      return out_node
  
    def _AbbreviatedTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Func')
      L = out_node.fields
      x0 = NewLeaf(self.name, color_e.StringConst)
      L.append(Field('name', x0))
  
      assert self.parsed is not None
      x1 = self.parsed.AbbreviatedTree()
      L.append(Field('parsed', x1))
  
      if self.pos_defaults is not None:  # List
        x2 = hnode.Array([])
        for i2 in self.pos_defaults:
          x2.children.append(i2.AbbreviatedTree())
        L.append(Field('pos_defaults', x2))
  
      if self.named_defaults is not None:  # Dict
        m = hnode.Leaf("Dict", color_e.OtherConst)
        x3 = hnode.Array([m])
        for k3, v3 in self.named_defaults.iteritems():
          x3.children.append(NewLeaf(k3, color_e.StringConst))
          x3.children.append(v3.AbbreviatedTree())
        L.append(Field('named_defaults', x3))
  
      if self.module_ is not None:  # Dict
        m = hnode.Leaf("Dict", color_e.OtherConst)
        x4 = hnode.Array([m])
        for k4, v4 in self.module_.iteritems():
          x4.children.append(NewLeaf(k4, color_e.StringConst))
          x4.children.append(v4.AbbreviatedTree())
        L.append(Field('module_', x4))
  
      return out_node
  
    def AbbreviatedTree(self):
      # type: () -> hnode_t
      return self._AbbreviatedTree()
  
  class Slice(value_t):
    _type_tag = 23
    __slots__ = ('lower', 'upper')
  
    def __init__(self, lower, upper):
      # type: (Optional[IntBox], Optional[IntBox]) -> None
      self.lower = lower
      self.upper = upper
  
    @staticmethod
    def CreateNull(alloc_lists=False):
      # type: () -> value.Slice
      return value.Slice(cast('Optional[IntBox]', None), cast('Optional[IntBox]', None))
  
    def PrettyTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Slice')
      L = out_node.fields
  
      if self.lower is not None:  # Optional
        x0 = self.lower.PrettyTree()
        L.append(Field('lower', x0))
  
      if self.upper is not None:  # Optional
        x1 = self.upper.PrettyTree()
        L.append(Field('upper', x1))
  
      return out_node
  
    def _AbbreviatedTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Slice')
      L = out_node.fields
      if self.lower is not None:  # Optional
        x0 = self.lower.AbbreviatedTree()
        L.append(Field('lower', x0))
  
      if self.upper is not None:  # Optional
        x1 = self.upper.AbbreviatedTree()
        L.append(Field('upper', x1))
  
      return out_node
  
    def AbbreviatedTree(self):
      # type: () -> hnode_t
      return self._AbbreviatedTree()
  
  class Range(value_t):
    _type_tag = 24
    __slots__ = ('lower', 'upper')
  
    def __init__(self, lower, upper):
      # type: (int, int) -> None
      self.lower = lower
      self.upper = upper
  
    @staticmethod
    def CreateNull(alloc_lists=False):
      # type: () -> value.Range
      return value.Range(-1, -1)
  
    def PrettyTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Range')
      L = out_node.fields
  
      x0 = hnode.Leaf(str(self.lower), color_e.OtherConst)
      L.append(Field('lower', x0))
  
      x1 = hnode.Leaf(str(self.upper), color_e.OtherConst)
      L.append(Field('upper', x1))
  
      return out_node
  
    def _AbbreviatedTree(self):
      # type: () -> hnode_t
      out_node = NewRecord('value.Range')
      L = out_node.fields
      x0 = hnode.Leaf(str(self.lower), color_e.OtherConst)
      L.append(Field('lower', x0))
  
      x1 = hnode.Leaf(str(self.upper), color_e.OtherConst)
      L.append(Field('upper', x1))
  
      return out_node
  
    def AbbreviatedTree(self):
      # type: () -> hnode_t
      return self._AbbreviatedTree()
  
  pass

class IntBox(pybase.CompoundObj):
  _type_tag = 64
  __slots__ = ('i',)

  def __init__(self, i):
    # type: (int) -> None
    self.i = i

  @staticmethod
  def CreateNull(alloc_lists=False):
    # type: () -> IntBox
    return IntBox(-1)

  def PrettyTree(self):
    # type: () -> hnode_t
    out_node = NewRecord('IntBox')
    L = out_node.fields

    x0 = hnode.Leaf(str(self.i), color_e.OtherConst)
    L.append(Field('i', x0))

    return out_node

  def _AbbreviatedTree(self):
    # type: () -> hnode_t
    out_node = NewRecord('IntBox')
    L = out_node.fields
    x0 = hnode.Leaf(str(self.i), color_e.OtherConst)
    L.append(Field('i', x0))

    return out_node

  def AbbreviatedTree(self):
    # type: () -> hnode_t
    return self._AbbreviatedTree()

class ProcDefaults(pybase.CompoundObj):
  _type_tag = 65
  __slots__ = ('for_word', 'for_typed', 'for_named')

  def __init__(self, for_word, for_typed, for_named):
    # type: (Optional[List[value_t]], Optional[List[value_t]], Optional[Dict[str, value_t]]) -> None
    self.for_word = for_word
    self.for_typed = for_typed
    self.for_named = for_named

  @staticmethod
  def CreateNull(alloc_lists=False):
    # type: () -> ProcDefaults
    return ProcDefaults(cast('Optional[List[value_t]]', None), cast('Optional[List[value_t]]', None), cast('Optional[Dict[str, value_t]]', None))

  def PrettyTree(self):
    # type: () -> hnode_t
    out_node = NewRecord('ProcDefaults')
    L = out_node.fields

    if self.for_word is not None:  # List
      x0 = hnode.Array([])
      for i0 in self.for_word:
        x0.children.append(i0.PrettyTree())
      L.append(Field('for_word', x0))

    if self.for_typed is not None:  # List
      x1 = hnode.Array([])
      for i1 in self.for_typed:
        x1.children.append(i1.PrettyTree())
      L.append(Field('for_typed', x1))

    if self.for_named is not None:  # Dict
      m = hnode.Leaf("Dict", color_e.OtherConst)
      x2 = hnode.Array([m])
      for k2, v2 in self.for_named.iteritems():
        x2.children.append(NewLeaf(k2, color_e.StringConst))
        x2.children.append(v2.PrettyTree())
      L.append(Field('for_named', x2))

    return out_node

  def _AbbreviatedTree(self):
    # type: () -> hnode_t
    out_node = NewRecord('ProcDefaults')
    L = out_node.fields
    if self.for_word is not None:  # List
      x0 = hnode.Array([])
      for i0 in self.for_word:
        x0.children.append(i0.AbbreviatedTree())
      L.append(Field('for_word', x0))

    if self.for_typed is not None:  # List
      x1 = hnode.Array([])
      for i1 in self.for_typed:
        x1.children.append(i1.AbbreviatedTree())
      L.append(Field('for_typed', x1))

    if self.for_named is not None:  # Dict
      m = hnode.Leaf("Dict", color_e.OtherConst)
      x2 = hnode.Array([m])
      for k2, v2 in self.for_named.iteritems():
        x2.children.append(NewLeaf(k2, color_e.StringConst))
        x2.children.append(v2.AbbreviatedTree())
      L.append(Field('for_named', x2))

    return out_node

  def AbbreviatedTree(self):
    # type: () -> hnode_t
    return self._AbbreviatedTree()

class LeftName(y_lvalue_t, sh_lvalue_t):
  _type_tag = 66
  __slots__ = ('name', 'blame_loc')

  def __init__(self, name, blame_loc):
    # type: (str, loc_t) -> None
    self.name = name
    self.blame_loc = blame_loc

  @staticmethod
  def CreateNull(alloc_lists=False):
    # type: () -> LeftName
    return LeftName('', cast(loc_t, None))

  def PrettyTree(self):
    # type: () -> hnode_t
    out_node = NewRecord('LeftName')
    L = out_node.fields

    x0 = NewLeaf(self.name, color_e.StringConst)
    L.append(Field('name', x0))

    assert self.blame_loc is not None
    x1 = self.blame_loc.PrettyTree()
    L.append(Field('blame_loc', x1))

    return out_node

  def _AbbreviatedTree(self):
    # type: () -> hnode_t
    out_node = NewRecord('LeftName')
    L = out_node.fields
    x0 = NewLeaf(self.name, color_e.StringConst)
    L.append(Field('name', x0))

    assert self.blame_loc is not None
    x1 = self.blame_loc.AbbreviatedTree()
    L.append(Field('blame_loc', x1))

    return out_node

  def AbbreviatedTree(self):
    # type: () -> hnode_t
    return self._AbbreviatedTree()

